#ifndef _MPEG2_QUANTISERSCALETABLE_H_
#define _MPEG2_QUANTISERSCALETABLE_H_

namespace MPEG2
{
	class CQuantiserScaleTable
	{
	public:
		static unsigned int		m_nTable0[0x20];
		static unsigned int		m_nTable1[0x20];
		
	};
}

#endif
