#ifndef _MPEG2_INVERSESCANTABLE_H_
#define _MPEG2_INVERSESCANTABLE_H_

namespace MPEG2
{
	class CInverseScanTable
	{
	public:
		static unsigned int		m_nTable0[0x40];
		static unsigned int		m_nTable1[0x40];
	};
}

#endif
